/*
 
 Copyright (c) 2011 Ethan Funk
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated 
 documentation files (the "Software"), to deal in the Software without restriction, including without limitation 
 the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, 
 and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in all copies or substantial portions 
 of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED 
 TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF 
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
 IN THE SOFTWARE.
 
*/

#include <stdio.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <string.h>
#include <sys/time.h>
#include <math.h>
#include "rs.h"
#include "rsp.h"

#define AUDIO_BLOCK_SIZE 128

unsigned char run;

void TERMhandler(int sig)
{
	run = 0;
}

void handleMetaData(cJSON *meta)
{
	char *metaStr;	

	if(metaStr = cJSON_PrintUnformatted(meta)){
		fprintf(stderr, "%s\n\n", metaStr);
		free(metaStr);
	}
	cJSON_Delete(meta);
}

void handleMessage(char *str)
{
	// note: str should not be freed
	fprintf(stderr, "{ \"server message\": \"%s\" }\n\n", str);
}

int main(int argc, const char * argv[])
{
	struct rspSession *rsp;
	unsigned char discont; 
	cJSON *root, *rspSection, *current;
	unsigned char err;

	run = 1;
	signal(SIGTERM, TERMhandler);
	signal(SIGINT, TERMhandler);
	signal(SIGABRT, TERMhandler);
	
	rsp = rspSessionNew("rspDecoder");
	// void* for you to set for your own reference
	rsp->referenceDesignator = NULL;					

	// get configuration information
	if(argc == 2){
		// Open the specified stream configuration file
		FILE *confFile = fopen(argv[1], "r");				
		if(confFile == NULL)
			return 1;
			
		if((rspSection = rspSessionReadConfigFile(confFile)) == NULL){
			fclose(confFile);										
			return 2;
		}
		fclose(confFile);											
		
	}else if(argc == 1){	
		// Read jSON format configuration from stdin
		size_t count, total;
		char *text = (char *)malloc(257);
		total = 0;
		while(1){
			count = fread(text+total, 1, 256, stdin);
			total = total + count;
			if(count == 256)
				text = realloc(text, total + 257);
			else
				break;
		}
		text[total] = 0;
		if((root = cJSON_Parse(text)) == NULL){
			free(text);
			return 2;
		}
		free(text);

		if((rspSection = cJSON_DetachItemFromObject(root, "rspStream")) == NULL){
			cJSON_Delete(root);
			return 2;
		}
		cJSON_Delete(root);
		
	}else
		// wrong number of passed arguments
		return 1;
	
	current = rspSection;
	// Try rspStream list entries loop, only if we are still in the running state	
	while(run){
		fprintf(stderr, "{\"status message\":\"Trying next rspStream list entry.\"}\n\n");
		if((err = rspSessionConfigNextJSON(rsp, &current)) == RSP_ERROR_END)
			// no more entries to try
			break;
		
		if(err != RSP_ERROR_NONE)
			// A problem with the current rspStream list entry... try another
			continue;
		
		rspSessionClear(rsp, TRUE);
		while(run){
			// try Network sockets and address records until there are no more IP4, IP6 or DNS records
			// for the current rspStream list entry, only if we are still in the running state.
			// Set socket time out to 10 seconds
			fprintf(stderr, "{\"status message\":\"Trying rspStream list entry address or next DNS record.\"}\n\n");
			if(rspSessionNextNetworkSetup(rsp, 10, NULL) != RSP_ERROR_NONE)
				// no more network records to try
				break;
			
			// !!!!! THIS IS IT:  THE MAIN PLAYING LOOP !!!!!
			
			// send a request to relay server or receiver report host to start sending stream packets... direct source will ignore this request
			if(rspPacketRecvrRequestSend(rsp, NULL, TRUE) != RSP_ERROR_NONE)
				continue;
			
			int size;
			unsigned char *data;
			unsigned char lastBlock, thisBlock, state;
			
			char *msg;
			cJSON *meta;
			discont = 1;
			data = NULL;
			
			lastBlock = 0;
			// state = 0, waiting to discover rsp format
			// state = 1, not waiting
			state = 0;		
			fprintf(stderr, "{\"status message\":\"Waiting to discover RSP format and stream Content format.\"}\n\n");
			while(run){
				// loop only if we are still in the running state
				size = rspSessionPlayTaskPush(rsp, &msg, &meta, &data, discont, -0.5);
				if((size == -3) || (size == -4)){
					// We have not received valid data in over 10 seconds OR socker error... handle!
					fprintf(stderr, "{\"status message\":\"Connection broken...\"}\n\n");
					break;			
				}
				if(!state && rsp->interleaver){
					fprintf(stderr, "{\"status message\":\"RSP format discovered. Decoder has been configured for %u bytes FEC, %u level interleaving and %u byte payloads with ", 
							rsp->FECroots, rsp->interleaving, rsp->colSize);
					if(rsp->flags & RSP_FLAG_CRC)
						fprintf(stderr, "packet CRC and "); 
					else
						fprintf(stderr, "no packet CRC and "); 
					if(rsp->flags & RSP_FLAG_RS)
						fprintf(stderr, "packet FEC."); 
					else
						fprintf(stderr, "no packet FEC."); 
					fprintf(stderr, "\"}\n\n");
					state = 1;
				}
				if(state && !rsp->interleaver){
					fprintf(stderr, "{\"status message\":\"RSP format change detected. Waiting to discover new RSP format.\"}\n\n");
					state = 0;
				}
				if(meta){
					// handle new meta data
					// check if this if this is stream content info...
					if(meta->child && meta->child->string && (strcasecmp(meta->child->string, "content") == 0)){
						// new stream content format info... 
						fprintf(stderr, "{\"status message\":\"Stream content format change detected. You may need to reconfigure down-stream decoders.\"}\n\n");
					}
					handleMetaData(meta);
				}		
				if(msg){
					// handle server message
					handleMessage(msg);
					free(msg);
				}
				if(size < 0){
					// Handle break in stream				
					if(discont == 0){
						fprintf(stderr, "{\"status message\":\"Filling interleaver...\"}\n\n");
					}
					discont = 1;
				}
				if(size > 0){
					// send out decoded stream data, and clear rebuffering flag
					discont = 0;
					write(STDOUT_FILENO, data, size);
				}
				
				// optional statistics section... executed after crossing a interleaver block boundry
				if((thisBlock = (unsigned char)floor(2.0 * rsp->avWrPosition)) != lastBlock){
					// display stats
					lastBlock = thisBlock;
					fprintf(stderr, "{\"status message\":\"Buf=%.0f, FEC=%.0f, ERR=%.0f, DUP=%.0f, BAD=%.0f\"}\n\n", rspSessionGetBalance(rsp) * 100.0, rsp->FECStat / rsp->FECroots * 100.0, rsp->ErrStat * 100.0, rsp->DupStat * 100.0, rsp->BadStat * 100.0);
				}
			}
			// send a request to relay server to stop sending stream packets
			rspPacketRecvrRequestSend(rsp, NULL, FALSE);
		}
		
	}
	if(run)
		fprintf(stderr, "{\"status message\":\"No more rspStream entries to try.\"}\n\n");
	else
		fprintf(stderr, "{\"status message\":\"Program has stopped.\"}\n\n");
	rspSessionFree(rsp);
}

